# $tmpl
#   --- однократные тесты на формат $tmpl
# + 1) шаблон
# + 2) файл
# +     - существующий
# +     - несуществующий
# +     - с правами (на файл/каталог) не дающими его читать
# +     - c \n в имени :)
# +     - некорректный формат шаблона (лишние @~ или ~@)
#   3) тесты для выполнения как отдельно, так и с \%attach и/или \%gpg
# +     - обычный шаблон
# +     - не содержит \n\n
#       - это шаблон MIME-письма, из двух частей (plain+html)
#       - это шаблон MIME-письма, из двух частей (plain+html), с картинкой
#         в html-части
#       - это шаблон а-ля пункт 3), но ещё и с приаттаченным файлом
# +     - в шаблоне есть Content-type с charset=koi8-r
# +     - шаблон в quoted-printable
# ?     - в шаблоне переводы строк \r\n
# \%data
#   --- однократные тесты с одним любым шаблоном
#   1) подстановка выполняется (в т.ч. для нескольких одинаковых @~...~@)
#   2) содержит лишние ключи
#   3) не содержит нужного ключа
#   4) вместо \%data undef; в шаблоне есть @~...~@
#   5) значение ключа: undef
#   6) значение ключа: q{}
# \%attach
#   --- однократные тесты на формат \%attach
#   1) undef (вероятно, все предыдущие тесты это уже многократно проверили :))
#   2) {}
#   3) ключ - q{}, значения - undef и строка
#   4) ключ - несуществующий файл
#   5) ключ - файл с правами (на файл/каталог) не дающими его читать
#   6) ключ - существующий файл с именем включающим кавычки и \n
#   7) ключ - существующий файл, значение - q{}
#   8) ключ - существующий файл, значение - "line1\nline2" (без \n в конце)
#   9) два ключа с одинаковыми именами (но разными путями к файлам)
#   10) multitest
# \%gpg
#   1) undef
#   2) {}
#   3) { encr_to => undef           }
#       проверить что берутся все адреса их To:, Cc:, Bcc:, folding, comments
#       проверить когда все адреса корректные, попадаются некорректные,
#           и нет ни одного адреса
#   4) { encr_to => []              }
#   5) { encr_to => ['wrong id']    }
#   6) { encr_to => ['good id1']    }
#   7) { encr_to => ['good id1', 'good id2']    }
#   8) { encr_to => ['good id1', 'wrong id']    }
#   9) { sign_by => undef           }
#   10){ sign_by => []              }
#   11){ sign_by => ['wrong id']    }
#   12){ sign_by => ['good id1']    }
#   13){ sign_by => ['good id1', 'good id2']    }
#   14){ sign_by => ['good id1', 'wrong id']    }
#   15){ encr_to => [], sign_by => []           }
#   16){ encr_to => ['good id1', 'good id2'], sign_by=>['good id3','good id4']}
#
# TODO multitest/crlf:
#       - $tmpl                 ^M в Subject
#       - \%attach              ^M в Subject и в последней строке body?!
#       - encr&sign             ^M в Subject
#       - attach+encr&sign      ^M в Subject
#       - sign                  ^M в Subject
#       - attach+sign           ^M в Subject и в последней строке body?!

use warnings;
use strict;

use Test::More;
use Test::Exception;
use Test::MockModule;

use File::Temp      qw( tempfile );

use POWER::Utils;
use POWER::GPG;

use POSIX qw(locale_h);
setlocale(LC_ALL, 'C');

use POWER::Email;

if ($> == 0) {
    plan skip_all => 'need non-root user';
    exit;
}
plan tests => 128;

system('chmod 0700 t/.gnupg');
# Prepare 'strange' files:
system('chmod 0 t/tmpl/not_readable*');
open my $f1, '>', "t/tmpl/line1\nline2.txt" or die "open: $!";
print {$f1} <<'EOF';
From: POWER Email <powerman-poweremail@powerman.asdfgroup.com>
To: @~to~@
Subject: @~subj~@

Hi! This is a test message.

@~msg~@

Bye!
EOF
close $f1;
open my $f2, '>', "t/tmpl/line1\"\nline2" or die "open: $!";
print {$f2} "test\n";
close $f2;
END {
    system('chmod 0700 t/tmpl/not_readable');
    system('chmod 0600 t/tmpl/not_readable.txt');
    unlink("t/tmpl/line1\nline2.txt");
    unlink("t/tmpl/line1\"\nline2");
}


gpg_config({ homedir => 't/.gnupg' });
my ($tmpl, %data, %attach, %gpg);
$tmpl           = Cat('t/tmpl/tmpl.txt');
$data{to}       = 'powerman@powerman.asdfgroup.com';
$data{msg}      = q{};


sub multitest {
    my ($test, $attach, $gpg) = @_;
    for my $file (glob('t/tmpl/multitest/*')) {
        (my $name = $file) =~ s!.*/|\.txt\z!!g;
        $data{subj} = "TEST: $test $name";
        lives_ok { Email($file, \%data, $attach, $gpg)  } $data{subj};
    }
}


my $mbox_fh;
sub mbox_open {
    die "mbox not closed" if $mbox_fh;
    open $mbox_fh, '<', shift or die "mbox_open: $!";
    return;
}
sub mbox_read {
    die "mbox not opened" if !$mbox_fh;
    my $from = <$mbox_fh>;
    die 'EOF' if !defined $from;
    die 'NO FROM' if $from !~ /\AFrom /;
    my @email;
    while (<$mbox_fh>) {
        if (!m/\AFrom /) {
            push @email, $_;
        }
        else {
            seek $mbox_fh, -length, 1;
            last;
        }
    }
    pop @email if $email[-1] =~ /\A\r?\n\z/;
    return join q{}, @email;
}
sub mbox_close {
    my $line = <$mbox_fh>;
    die 'NO EOF' if defined $line;
    close $mbox_fh;
    undef $mbox_fh;
    return;
}


my $m = new Test::MockModule('POWER::Email');
$m->mock('_send_mail', sub {
    my $mail = shift;
    my $wait = mbox_read();
    ### NORMALIZE $mail & $wait to be able to compare them:
    # strip transport/mbox-related fields
    $wait =~ s/^(?:Return-Path|Delivered-To|Received|Date|Message-ID|Status|Content-Length|Lines):.*\n//gmi;
    # CRLF become LF in From&To fields while delivering?!
    $mail =~ s/^((?:From|To):.*)\r\n/$1\n/gmi;
    # Bcc field doesn't delivered
    $mail =~ s/^(?:Bcc):.*\n//gmi;
    for ($wait, $mail) {
        # field folding may change while delivering
        my ($head, $body) = POWER::Email::_split2headbody($_);
        $head =~ s/\s*\n[ \t]+/ /g;
        $_ = $head."\n".$body;
        # boundary is random
        if (m{^Content-Type:.*;\s+BOUNDARY="([^"]+)"}mig) {
            s/\Q$1/_b_o_u_n_d_a_r_y_/g;
        }
        # decrypt encrypted and encrypted&signed messages
        s[(-----BEGIN PGP MESSAGE-----.*?-----END PGP MESSAGE-----\s*)]
         [  my ($plain, $sign) = eval { gpg_decrypt($1) };
            $sign = join '\n', map {"$_ => $sign->{$_}"} sort keys %$sign;
            join ":::::\n", $@ ? $@ : ($plain, $sign)
         ]gse;
        # verify signed messages
        s[(--_b_o_u_n_d_a_r_y_\n(.*?)\n--_b_o_u_n_d_a_r_y_\n.*?)(-----BEGIN PGP SIGNATURE-----.*?-----END PGP SIGNATURE-----)]
         [  my ($sign) = eval { gpg_verify($2, $3) };
            $sign = join '\n', map {"$_ => $sign->{$_}"} sort keys %$sign;
            $1 . ($@ ? $@ : $sign)
         ]gse;
        # after verifying signature fix second-level random boundary
        if (m{^Content-Type:.*;\s+BOUNDARY="(?!_b_o_u_n_d_a_r_y_)([^"]+)"}mig) {
            s/\Q$1/_b_o_u_n_d_a_r_y_2_/g;
        }
    }
    is($mail, $wait);
#    Echo('> /tmp/mail', $mail); Echo('> /tmp/wait', $wait); exit if $mail ne $wait;
    return;
});

###
### $tmpl
###

mbox_open('t/mbox/tmpl');

$data{subj}     = 'TEST: $tmpl (1) template';
lives_ok  { Email($tmpl, \%data)                    } $data{subj};

$data{subj}     = 'TEST: $tmpl (2) file, exists';
lives_ok  { Email('t/tmpl/tmpl.txt', \%data)        } $data{subj};

throws_ok { Email(q{})                              } qr/open: No such file/;
throws_ok { Email('t/tmpl/NO_SUCH.txt')             } qr/open: No such file/;
throws_ok { Email('t/tmpl/not_readable.txt')        } qr/open: Permission denied/;
throws_ok { Email('t/tmpl/not_readable/tmpl.txt')   } qr/open: Permission denied/;

# file with \n in name processed as template
throws_ok { Email("t/tmpl/line1\nline2.txt")        } qr{no head/body separator};

throws_ok { Email('t/tmpl/tmpl_wrong1.txt', \%data) } qr/no value for @~@~msg~@/;
throws_ok { Email('t/tmpl/tmpl_wrong2.txt', \%data) } qr/orphaned ~@/;
throws_ok { Email('t/tmpl/tmpl_wrong3.txt', \%data) } qr/orphaned @~/;
throws_ok { Email('t/tmpl/tmpl_wrong4.txt', \%data) } qr/wrong template: @~~@/;
throws_ok { Email('t/tmpl/tmpl_wrong5.txt', \%data) } qr/no value for @~\nmsg\n~@/;

multitest('$tmpl (3)');

mbox_close();

###
### %data
###

mbox_open('t/mbox/data');

$data{msg}      = "line1\nline2";
$data{subj}     = 'TEST: \%data (1) @~msg~@ @~msg~@';
lives_ok  { Email($tmpl, \%data)                    } $data{subj};

$data{unused}   = "unused";
$data{subj}     = 'TEST: \%data (2) unused key';
lives_ok  { Email($tmpl, \%data)                    } $data{subj};

delete $data{msg};
throws_ok { Email($tmpl, \%data)                    } qr/no value for @~msg~@/;

throws_ok { Email($tmpl)                            } qr/no value for @~to~@/;

$data{msg}      = undef;
throws_ok { Email($tmpl, \%data)                    } qr/no value for @~msg~@/;

$data{msg}      = q{};
$data{subj}     = 'TEST: \%data (6) {msg}=""';
lives_ok  { Email($tmpl, \%data)                    } $data{subj};

mbox_close();

###
### \%attach
###

mbox_open('t/mbox/attach');

$data{subj}     = 'TEST: \%attach (2) {}';
lives_ok  { Email($tmpl, \%data, {})                } $data{subj};

throws_ok { Email($tmpl, \%data, {q{} => undef})    } qr/No such file/;

$data{subj}     = 'TEST: \%attach (3) {""=>"test"}';
lives_ok  { Email($tmpl, \%data, {q{} => 'test'})   } $data{subj};

throws_ok { Email($tmpl, \%data, {'nosuchfile'=>undef}) } qr/No such file/;

throws_ok { Email($tmpl, \%data, {'t/tmpl/not_readable.txt'=>undef})} qr/Permission denied/;

throws_ok { Email($tmpl, \%data, {'t/tmpl/not_readable/tmpl.txt'=>undef})} qr/Permission denied/;

$data{subj}     = 'TEST: \%attach (6) {line1"\nline2=>undef}';
lives_ok  { Email($tmpl, \%data, {"t/tmpl/line1\"\nline2"=>undef})} $data{subj};

$data{subj}     = 'TEST: \%attach (7) {tmpl.txt => ""}';
lives_ok  { Email($tmpl, \%data, {'t/tmpl/tmpl.txt'=>q{}})} $data{subj};

$data{subj}     = 'TEST: \%attach (8) {tmpl.txt => "line1\nline2"}';
lives_ok  { Email($tmpl, \%data, {'t/tmpl/tmpl.txt'=>"line1\nline2"})} $data{subj};

$data{subj}     = 'TEST: \%attach (9) {tmpl.txt => undef, tmpl.txt => undef}';
lives_ok  { Email($tmpl, \%data, {
    't/tmpl/tmpl.txt'           => undef,
    't/tmpl/../tmpl/tmpl.txt'   => undef,
    })} $data{subj};

multitest('\%attach (10)', {});

mbox_close();

###
### \%gpg
###

#$m->unmock_all();

mbox_open('t/mbox/gpg');

$data{subj}     = q{};
throws_ok { Email($tmpl, \%data, undef, {})                 }
    qr/{encr_to} or {sign_by}/;

throws_ok { Email($tmpl, \%data, undef, {encr_to=>undef})   }
    qr/{encr_to} undefined/;

$data{subj}     = 'TEST: \%gpg (4) {encr_to=>[]}';
lives_ok  { Email($tmpl, \%data, undef, {encr_to=>[]})      } $data{subj};

$data{subj}     = 'TEST: \%gpg (4) {encr_to=>[]} +bcc';
lives_ok  { Email('t/tmpl/tmpl_bcc.txt',
                         \%data, undef, {encr_to=>[]})      } $data{subj};

$data{to}       = 'powerman@betspider.net';
$data{subj}     = 'TEST: \%gpg (4) {encr_to=>[]} manyrcp';
lives_ok  { Email('t/tmpl/tmpl_manyrcp.txt', \%data, undef, {encr_to=>[]})
                                                            } $data{subj};
$data{to}       = 'nosuch@key.com';
throws_ok { Email('t/tmpl/tmpl_manyrcp.txt', \%data, undef, {encr_to=>[]}) }
    qr/nosuch\@key\.com: skipped: no public key/i;

$data{to}       = q{};
throws_ok { Email('t/tmpl/tmpl_norcp.txt', \%data, undef, {encr_to=>[]})    }
    qr/recipients list is empty/;

$data{to}       = 'powerman@powerman.asdfgroup.com';
throws_ok { Email($tmpl, \%data, undef, {encr_to=>['nosuch@key.com']})   }
    qr/nosuch\@key\.com: skipped: no public key/i;

$data{subj}     = 'TEST: \%gpg (6) {encr_to=>["POWER Email"]}';
lives_ok  { Email($tmpl, \%data, undef, {encr_to=>['POWER Email']}) } $data{subj};

$data{subj}     = 'TEST: \%gpg (7) {encr_to=>["POWER Email","Efros"]}';
lives_ok  { Email($tmpl, \%data, undef, {encr_to=>['POWER Email','Efros']}) } $data{subj};

throws_ok { Email($tmpl, \%data, undef, {encr_to=>['POWER Email','nosuch@key.com']}) }
    qr/nosuch\@key\.com: skipped: no public key/i;

$data{subj}     = q{};
throws_ok { Email($tmpl, \%data, undef, {sign_by=>undef})   }
    qr/{sign_by} undefined/;

$data{subj}     = 'TEST: \%gpg (10) {sign_by=>[]}';
lives_ok  { Email($tmpl, \%data, undef, {sign_by=>[]})      } $data{subj};

throws_ok { Email($tmpl, \%data, undef, {sign_by=>['nosuch@key.com']})   }
    qr/"nosuch\@key\.com": no secret key/i;

$data{subj}     = 'TEST: \%gpg (12) {sign_by=>["POWER Email"]}';
lives_ok  { Email($tmpl, \%data, undef, {sign_by=>['POWER Email']}) } $data{subj};

$data{subj}     = 'TEST: \%gpg (13) {sign_by=>["POWER Email","Baba"]}';
lives_ok  { Email($tmpl, \%data, undef, {sign_by=>['POWER Email','Baba']}) } $data{subj};

throws_ok { Email($tmpl, \%data, undef, {sign_by=>['Baba','nosuch@key.com']})   }
    qr/"nosuch\@key\.com": no secret key/i;

$data{subj}     = 'TEST: \%gpg (15) {encr_to=>[], sign_by=>[]}';
lives_ok  { Email($tmpl, \%data, undef, {encr_to=>[],sign_by=>[]})} $data{subj};

$data{subj}     = 'TEST: \%gpg (16) {encr_to=>["POWER Email","Efros"], sign_by=>["POWER Email","Baba"]}';
lives_ok  { Email($tmpl, \%data, undef, {
    encr_to=>['POWER Email','Efros'], sign_by=>['POWER Email','Baba']
    })} $data{subj};

multitest('\%gpg () encr&sign', undef, {encr_to=>[],sign_by=>[]});
multitest('\%gpg () attach+encr&sign',
    { 't/tmpl/tmpl.txt'  => undef },
    { encr_to => [], sign_by => [] }
    );
multitest('\%gpg () sign', undef, {sign_by=>[]});
multitest('\%gpg () attach+sign',
    { 't/tmpl/tmpl.txt'  => undef },
    { sign_by => [] }
    );

mbox_close();

