use warnings;
use strict;

use Test::More;
use Test::Exception;
use Test::MockModule;

use POWER::Email;

plan tests => 16;


my $Mail;
my $m = new Test::MockModule('POWER::Email');
$m->mock('_send_mail', sub { $Mail = shift; return });

my $tmpl;
my $C_T = "Content-Type: TEXT/PLAIN; charset=US-ASCII\n";

throws_ok { Email()                     }   qr/open:/;
throws_ok { Email(undef)                }   qr/open:/;
throws_ok { Email(q{})                  }   qr/open:/;


throws_ok { Email('/dev/null')          }   qr{no head/body separator found};


$tmpl = <<'EO_TMPL';
To: root
Content-Type: TEXT/PLAIN; charset=US-ASCII
EO_TMPL
throws_ok { Email($tmpl)                }   qr{no head/body separator found};
Email($tmpl."\n");
is($Mail, $tmpl."\n\n",                     'empty body replaced with \n');
Email($tmpl."\n\n");
is($Mail, $tmpl."\n\n",                     '\n body not modified');
Email($tmpl."\n\n\n");
is($Mail, $tmpl."\n\n\n",                   '\n\n body not modified');


$tmpl = <<'EO_TMPL';
To: root
Content-Type: text/html; charset=koi8-r

EO_TMPL
Email($tmpl);
is($Mail, $tmpl."\n",                       'user Content-Type');


$tmpl = <<'EO_TMPL';
To: root


EO_TMPL
Email($tmpl);
is($Mail, "To: root\n$C_T\n\n",               'default Content-Type');


$tmpl = <<'EO_TMPL';
To: root@localhost
Content-Type: TEXT/PLAIN; charset=US-ASCII

EO_TMPL
Email($tmpl);
is($Mail, $tmpl."\n",                       'simple static template');
Email('t/tmpl/smoke.tmpl');
is($Mail, $tmpl."\n",                       'simple static template in file');


$tmpl = <<'EO_TMPL';
To: @~email~@
Content-Type: TEXT/PLAIN; charset=US-ASCII

EO_TMPL
Email($tmpl, {email=>'nobody@localhost'});
is($Mail, "To: nobody\@localhost\n$C_T\n\n",'simple dynamic template');


Email('t/tmpl/smoke.tmpl', undef, {'t/tmpl/smoke.tmpl' => undef});
like($Mail, qr{MULTIPART/MIXED.*filename="smoke\.tmpl"}is,
                                            'attach file');


throws_ok { Email('t/tmpl/smoke.tmpl', undef, {'no_such' => undef})} qr/open:/;
Email('t/tmpl/smoke.tmpl', undef, {'no_such' => 'fake'});
like($Mail, qr{MULTIPART/MIXED.*filename="no_such"}is,
                                            'attach scalar');

